// Generated by CoffeeScript 1.9.1
(function() {
  var ALLOWED_TAG_NAMES, appendFootnoteDefinitionCollection, applyFootnoteNumber, convertPreToRawHTML, convertScatteredPreToRawHTML, createNodeFromHTMLFragment, decomposeHTMLNode, decomposeHTMLNodes, decomposeHTMLString, defineFootnoteNumber, endIndexFromPosition, foldHTMLNodes, isVoidElement, preprocess, removeDefinitions, sanitizeTag, startIndexFromPosition, wrapHTMLNodeInParagraph,
    indexOf = [].indexOf || function(item) { for (var i = 0, l = this.length; i < l; i++) { if (i in this && this[i] === item) return i; } return -1; };

  preprocess = function(root, sourceText, options) {
    var defs, mapping;
    convertPreToRawHTML(root);
    convertScatteredPreToRawHTML(root, sourceText);
    root.children = decomposeHTMLNodes(root.children);
    root.children = foldHTMLNodes(root.children);
    if (options.footnotes) {
      mapping = defineFootnoteNumber(root).mapping;
      applyFootnoteNumber(root, mapping);
    }
    defs = removeDefinitions(root);
    if (options.footnotes) {
      appendFootnoteDefinitionCollection(root, defs);
    }
    root = wrapHTMLNodeInParagraph(root);
    root = sanitizeTag(root);
    return [root, defs];
  };

  defineFootnoteNumber = function(node, num, mapping) {
    var child, id, j, len1, ref;
    if (num == null) {
      num = 1;
    }
    if (mapping == null) {
      mapping = {};
    }
    ref = node.children;
    for (j = 0, len1 = ref.length; j < len1; j++) {
      child = ref[j];
      if (child.type === 'footnoteReference') {
        id = child.identifier;
        if (mapping[id] == null) {
          mapping[id] = num;
          num += 1;
        }
        child.footnoteNumber = mapping[id];
      }
      if (child.children) {
        num = defineFootnoteNumber(child, num, mapping).maxNumber;
      }
    }
    return {
      mapping: mapping,
      maxNumber: num
    };
  };

  applyFootnoteNumber = function(node, mapping) {
    var child, id, isFootnoteDefLike, j, len1, ref, results;
    if (node.children == null) {
      return;
    }
    ref = node.children;
    results = [];
    for (j = 0, len1 = ref.length; j < len1; j++) {
      child = ref[j];
      isFootnoteDefLike = child.type === 'definition' && /^[^]/.test(child.identifier);
      if (child.type === 'footnoteDefinition' || isFootnoteDefLike) {
        id = isFootnoteDefLike ? child.identifier.slice(1) : child.identifier;
        child.footnoteNumber = mapping[id] || 0;
      }
      results.push(applyFootnoteNumber(child, mapping));
    }
    return results;
  };

  appendFootnoteDefinitionCollection = function(node, defs) {
    var def, footnoteDefs;
    footnoteDefs = (function() {
      var j, len1, results;
      results = [];
      for (j = 0, len1 = defs.length; j < len1; j++) {
        def = defs[j];
        if ((def.footnoteNumber != null) && def.footnoteNumber > 0) {
          results.push(def);
        }
      }
      return results;
    })();
    footnoteDefs.sort(function(a, b) {
      return a.footnoteNumber - b.footnoteNumber;
    });
    if (footnoteDefs.length > 0) {
      return node.children.push({
        type: 'footnoteDefinitionCollection',
        children: footnoteDefs
      });
    }
  };

  removeDefinitions = function(node) {
    var child, childDefs, children, defs, j, len1, ref, ref1;
    if (node.children == null) {
      return [];
    }
    children = [];
    defs = [];
    ref = node.children;
    for (j = 0, len1 = ref.length; j < len1; j++) {
      child = ref[j];
      if ((ref1 = child.type) === 'definition' || ref1 === 'footnoteDefinition') {
        defs.push(child);
      } else {
        childDefs = removeDefinitions(child);
        Array.prototype.push.apply(defs, childDefs);
        children.push(child);
      }
    }
    node.children = children;
    return defs;
  };

  convertPreToRawHTML = function(root) {
    var j, len1, node, ref, results;
    ref = root.children;
    results = [];
    for (j = 0, len1 = ref.length; j < len1; j++) {
      node = ref[j];
      if (node.type === 'html' && /^<pre[ >][^]*<\/pre>$/i.test(node.value)) {
        results.push(node.subtype = 'raw');
      } else {
        results.push(void 0);
      }
    }
    return results;
  };

  convertScatteredPreToRawHTML = function(root, sourceText) {
    var endPreNode, i, isEnd, isStart, j, k, len1, len2, node, offset, paraLastNode, pre, preTexts, rawHTML, rawHTMLNode, ref, results, sliceEnd, sliceStart, sourceLines, start, startParaIndex, startPreNode;
    preTexts = [];
    startPreNode = null;
    startParaIndex = null;
    sourceLines = null;
    ref = root.children;
    for (i = j = 0, len1 = ref.length; j < len1; i = ++j) {
      node = ref[i];
      isStart = node.type === 'html' && /^<pre[ >]/i.test(node.value);
      if (isStart) {
        startPreNode = node;
        startParaIndex = i;
      }
      paraLastNode = null;
      isEnd = ((startPreNode != null) && node.type === 'html' && /<\/pre>$/i.test(node.value)) || ((startPreNode != null) && node.type === 'paragraph' && (paraLastNode = node.children[node.children.length - 1]) && paraLastNode.type === 'html' && /<\/pre>$/i.test(paraLastNode.value));
      if (isEnd) {
        endPreNode = paraLastNode != null ? paraLastNode : node;
        if (sourceLines == null) {
          sourceLines = sourceText.split(/^/m);
        }
        sliceStart = startIndexFromPosition(startPreNode.position, sourceLines);
        sliceEnd = endIndexFromPosition(endPreNode.position, sourceLines);
        rawHTML = sourceText.slice(sliceStart, sliceEnd);
        preTexts.push({
          startParaIndex: startParaIndex,
          paraCount: i - startParaIndex + 1,
          rawHTML: rawHTML
        });
        startPreNode = null;
        startParaIndex = null;
      }
    }
    offset = 0;
    results = [];
    for (k = 0, len2 = preTexts.length; k < len2; k++) {
      pre = preTexts[k];
      rawHTMLNode = {
        type: 'html',
        subtype: 'raw',
        value: pre.rawHTML
      };
      start = pre.startParaIndex - offset;
      root.children.splice(start, pre.paraCount, rawHTMLNode);
      results.push(offset = pre.paraCount - 1);
    }
    return results;
  };

  startIndexFromPosition = function(pos, lines) {
    var i, index, j, ref;
    index = 0;
    for (i = j = 0, ref = pos.start.line - 1; 0 <= ref ? j < ref : j > ref; i = 0 <= ref ? ++j : --j) {
      index += lines[i].length;
    }
    index += pos.start.column - 1;
    return index;
  };

  endIndexFromPosition = function(pos, lines) {
    var i, index, j, ref;
    index = 0;
    for (i = j = 0, ref = pos.end.line - 1; 0 <= ref ? j < ref : j > ref; i = 0 <= ref ? ++j : --j) {
      index += lines[i].length;
    }
    index += pos.end.column - 1;
    return index;
  };

  foldHTMLNodes = function(nodes) {
    var children, folded, index, j, k, len1, len2, node, pNode, processedNodes, startTag, startTagIndex;
    processedNodes = [];
    for (j = 0, len1 = nodes.length; j < len1; j++) {
      node = nodes[j];
      if (node.subtype === 'end') {
        startTagIndex = null;
        for (index = k = 0, len2 = processedNodes.length; k < len2; index = ++k) {
          pNode = processedNodes[index];
          if (pNode.subtype === 'start' && pNode.tagName === node.tagName) {
            startTagIndex = index;
          }
        }
        if (startTagIndex == null) {
          processedNodes.push(node);
          continue;
        }
        startTag = processedNodes[startTagIndex];
        children = processedNodes.splice(startTagIndex).slice(1);
        folded = {
          type: 'html',
          subtype: 'folded',
          tagName: startTag.tagName,
          startTag: startTag,
          endTag: node,
          children: children
        };
        processedNodes.push(folded);
      } else {
        if (node.children != null) {
          node.children = foldHTMLNodes(node.children);
        }
        processedNodes.push(node);
      }
    }
    return processedNodes;
  };

  decomposeHTMLNodes = function(nodes) {
    var fragmentNodes, j, len1, node, processedNodes;
    processedNodes = [];
    for (j = 0, len1 = nodes.length; j < len1; j++) {
      node = nodes[j];
      if (node.type === 'html' && node.subtype === 'raw') {
        processedNodes.push(node);
      } else if (node.type === 'html') {
        fragmentNodes = decomposeHTMLNode(node);
        if (fragmentNodes != null) {
          processedNodes.push.apply(processedNodes, fragmentNodes);
        } else {
          node.subtype = 'malformed';
          processedNodes.push(node);
        }
      } else {
        if (node.children != null) {
          node.children = decomposeHTMLNodes(node.children);
        }
        processedNodes.push(node);
      }
    }
    return processedNodes;
  };

  decomposeHTMLNode = function(node) {
    var fragments, value;
    value = node.value;
    if (node.position.start.line === node.position.end.line) {
      value = value.replace(/\n\n/, '');
    }
    fragments = decomposeHTMLString(value);
    return fragments != null ? fragments.map(createNodeFromHTMLFragment) : void 0;
  };

  decomposeHTMLString = function(str) {
    var matches, sumLength;
    if (str === '') {
      return null;
    }
    matches = str.match(/<[^>]*>|[^<>]+/g);
    sumLength = matches.reduce((function(len, s) {
      return len + s.length;
    }), 0);
    if (sumLength !== str.length) {
      return null;
    } else {
      return matches;
    }
  };

  createNodeFromHTMLFragment = function(str) {
    var j, name, ref, ref1, slash, subtype;
    if (/^[^<]/.test(str)) {
      return {
        type: 'text',
        value: str
      };
    }
    ref1 = (ref = /^<(\/?)([0-9A-Z]+)/i.exec(str)) != null ? ref : [], j = ref1.length - 2, slash = ref1[j++], name = ref1[j++];
    subtype = name == null ? 'special' : slash === '/' ? 'end' : isVoidElement(name) ? 'void' : 'start';
    return {
      type: 'html',
      subtype: subtype,
      tagName: name,
      value: str
    };
  };

  isVoidElement = function(elementName) {
    var voidElementNames;
    voidElementNames = ['area', 'base', 'br', 'col', 'command', 'embed', 'hr', 'img', 'input', 'keygen', 'link', 'meta', 'param', 'source', 'track', 'wbr'];
    return voidElementNames.indexOf(elementName) !== -1;
  };

  wrapHTMLNodeInParagraph = function(root) {
    var child, children, j, len1, ref;
    children = [];
    ref = root.children;
    for (j = 0, len1 = ref.length; j < len1; j++) {
      child = ref[j];
      if (child.type === 'html') {
        children.push({
          type: 'paragraph',
          children: [child]
        });
      } else {
        children.push(child);
      }
    }
    root.children = children;
    return root;
  };

  ALLOWED_TAG_NAMES = ['a', 'abbr', 'b', 'br', 'cite', 'code', 'del', 'dfn', 'em', 'i', 'img', 'input', 'ins', 'kbd', 'mark', 'ruby', 'rp', 'rt', 'q', 's', 'samp', 'small', 'span', 'strong', 'sub', 'sup', 'u', 'wbr'];

  sanitizeTag = function(node) {
    var child, children, j, len1, ref, ref1;
    if (node.children == null) {
      return node;
    }
    children = [];
    ref = node.children;
    for (j = 0, len1 = ref.length; j < len1; j++) {
      child = ref[j];
      if (child.subtype === 'folded' && (ref1 = child.tagName, indexOf.call(ALLOWED_TAG_NAMES, ref1) < 0)) {
        children.push(child.startTag);
        Array.prototype.push.apply(children, sanitizeTag(child).children);
        children.push(child.endTag);
      } else {
        children.push(sanitizeTag(child));
      }
    }
    node.children = children;
    return node;
  };

  module.exports = preprocess;

}).call(this);
